'''
Created on 25 apr 2019

@author: Matteo
'''
# Imports
import logging
import asyncio
from . import _LOGGER
from .const import (CD_RETURN_IMMEDIATELY,CD_CONTINUE_WAITING)
from .orvibo_udp import (OrviboUDP,MAGIC,PADDING_1,DISCOVERY_S20,PADDING_2)

STATECHANGE_ID  = b'\x64\x63'
STATECHANGE_LEN = b'\x00\x17'
# Datagram protocol

class S20(OrviboUDP):

    def __init__(self,hp,mac,mytime = None,timeout=3,**kwargs):
        OrviboUDP.__init__(self, hp, mac, mytime, timeout, ** kwargs)
        self.state = -1

    def use_subscribe_data(self,s_data):
        self.state = 0 if s_data[-1:]==b'\x00' else 1
        
    def check_statechange_packet(self,data,addr):
        return CD_RETURN_IMMEDIATELY if len(data)>6 and data[4:6] == (STATECHANGE_ID) and self.is_my_mac(data) else CD_CONTINUE_WAITING
    
        
    async def state_change(self,newst,timeout=-1,retry=3):
        if await self.subscribe_if_necessary():
            newst = 1 if int(newst) else 0
            pkt = MAGIC + STATECHANGE_LEN + STATECHANGE_ID + self.mac + PADDING_1\
                + PADDING_2+(b'\x01' if newst else b'\x00')
            timeout = self.timeout if timeout<=0 else timeout
            if await OrviboUDP.protocol(pkt,self.hp,self.check_statechange_packet,timeout,retry):
                self.state = newst
                return True
        return False
    
    @staticmethod
    async def discovery(broadcast_address='255.255.255.255',timeout=5,retry=3):
        disc = await OrviboUDP.discovery(broadcast_address, timeout, retry)
        hosts = dict()
        for k,v in disc.items():
            if v['type']==DISCOVERY_S20:
                s = S20(**v)
                s.state = 0 if v['raw'][-1:]==b'\x00' else 1
                hosts[k] = s
        return hosts

if __name__ == '__main__': # pragma: no cover
    import sys
    import traceback
    async def testFake(n):
        for i in range(n):
            _LOGGER.debug("Counter is %d",i)
            await asyncio.sleep(1)
    async def discoveryTest():
        v = await S20.discovery('192.168.25.255', 7, 3)
        if v:
            _LOGGER.info("Discovery str %s",v)
        else:
            _LOGGER.warning("Discovery failed")
    
    async def set_state_test(st):
        a = S20(('192.168.25.42',10000),b'\xac\xcf\x23\x93\x34\x9c')
        rv = await a.state_change(st)
        if rv:
            _LOGGER.info("Change state ok")
        else:
            _LOGGER.warning("Change state failed")
    _LOGGER.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    _LOGGER.addHandler(handler)
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(discoveryTest())
        asyncio.ensure_future(testFake(10))
        loop.run_until_complete(set_state_test(int(sys.argv[1])))
    except BaseException as ex:
        _LOGGER.error("Test error %s",str(ex))
    except:
        traceback.print_exc()
    finally:
        loop.close()


'''
Created on 25 apr 2019

@author: Matteo
'''
import logging
import binascii
import struct
import asyncio
from random import randint
import time
from . import _LOGGER
from .const import (CD_RETURN_IMMEDIATELY,CD_CONTINUE_WAITING)
from .orvibo_udp import (OrviboUDP,MAGIC,PADDING_1,DISCOVERY_ALLONE)

LEARNIR_ID  = b'\x6c\x73'
LEARNIR_LEN = b'\x00\x18'
LEARNIR_2 = b'\x01\x00\x00\x00\x00\x00'
EMITIR_ID  = b'\x69\x63'
EMITIR_2 = b'\x65\x00\x00\x00'

LEARN_MAX_TIME = 40
# Datagram protocol

class AllOne(OrviboUDP):

            
    def check_emitir_packet(self,data,addr):
        return CD_RETURN_IMMEDIATELY if len(data)>=6 and data[4:6] == (EMITIR_ID) and self.is_my_mac(data) else CD_CONTINUE_WAITING
        
    async def emit_ir(self,irc,timeout=-1,retry=3):
        if await self.subscribe_if_necessary():
            plen = struct.pack('>H',len(irc)+26)
            ilen = struct.pack('<H',len(irc))
            rnd = struct.pack('<H',randint(0,65535))
            pkt = MAGIC + plen + EMITIR_ID + self.mac + PADDING_1\
                + EMITIR_2 + rnd + ilen + irc
            timeout = self.timeout if timeout<=0 else timeout
            rv = await OrviboUDP.protocol(pkt,self.hp,self.check_emitir_packet,timeout,retry)
            if rv:
                return rv[0]
        return None
        
    def check_learnir_init_packet(self,data,addr):
        return CD_RETURN_IMMEDIATELY if len(data)>=6 and data[4:6] == (LEARNIR_ID) and data[2:4]==LEARNIR_LEN and self.is_my_mac(data) else CD_CONTINUE_WAITING
    
    def check_learnir_get_packet(self,data,addr):
        return CD_RETURN_IMMEDIATELY if len(data)>=6 and data[4:6] == (LEARNIR_ID) and data[2:4]>LEARNIR_LEN and self.is_my_mac(data) else CD_CONTINUE_WAITING
        
    async def learn_ir_init(self,timeout=-1,retry=3):
        if await self.subscribe_if_necessary():
            pkt = MAGIC + LEARNIR_LEN + LEARNIR_ID + self.mac + PADDING_1\
                    + LEARNIR_2
            timeout = self.timeout if timeout<=0 else timeout
            if await OrviboUDP.protocol(pkt,self.hp,self.check_learnir_init_packet,timeout,retry):
                self.learning_time = time.time()
                return True
        return False
    
    async def learn_ir_get(self,timeout=30):
        to = min(LEARN_MAX_TIME-(time.time()-self.learning_time),timeout)
        if to>0:
            rv = await OrviboUDP.protocol(None, self.hp,self.check_learnir_get_packet,to,1)
            if rv and len(rv[0])>26:
                return rv[0][26:]
        return None
    
    @staticmethod
    async def discovery(broadcast_address='255.255.255.255',timeout=5,retry=3):
        disc = await OrviboUDP.discovery(broadcast_address, timeout, retry)
        hosts = dict()
        for k,v in disc.items():
            if v['type']==DISCOVERY_ALLONE:
                hosts[k] = AllOne(**v)
        return hosts

if __name__ == '__main__': # pragma: no cover
    import sys
    import traceback
    async def testFake(n):
        for i in range(n):
            _LOGGER.debug("Counter is %d",i)
            await asyncio.sleep(1)
    async def discoveryTest():
        v = await AllOne.discovery('192.168.25.255', 7, 3)
        if v:
            _LOGGER.info("Discovery str %s",v)
        else:
            _LOGGER.warning("Discovery failed")
    
    async def emit_test(keystr):
        payload = binascii.unhexlify(keystr)
        a = AllOne(('192.168.25.41',10000),b'\xac\xcf\x23\x72\x5a\x50')
        rv = await a.emit_ir(payload)
        if rv:
            _LOGGER.info("Emit OK %s",binascii.hexlify(rv).decode('utf-8'))
        else:
            _LOGGER.warning("Emit failed")
    async def learn_test():
        a = AllOne(('192.168.25.41',10000),b'\xac\xcf\x23\x72\x5a\x50')
        rv = await a.learn_ir_init()
        if rv:
            _LOGGER.info("Entered learning mode: please press key")
            rv = await a.learn_ir_get()
            if rv:
                _LOGGER.info("Obtained %s",binascii.hexlify(rv).decode('utf-8'))
            else:
                _LOGGER.warning("No key pressed")
        else:
            _LOGGER.warning("Enter learning failed")
    _LOGGER.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    _LOGGER.addHandler(handler)
    loop = asyncio.get_event_loop()
    try:
        #asyncio.ensure_future(testFake(10))
        #loop.run_until_complete(emit_test('00000000a801000000000000000098018e11951127029b0625029906270299062702380227023a0225023802270238022d023202270299062702990627029806270238022702380227023802270238022802370227023802270238022702980627023802240245021c02380227023802270238022702980627029c0623023802270298062702990627029b062502990627029906270220b7a1119d11270299062702990628029b06250238022702380227023802270238022702380227029906270299062702990627023802270238022a0234022702380227023802260238022702380226029a06260238022602380226023802260241021e02380227029b0624029906270238022702980627029b0625029906270299062702990629021db79f11a2112502990627029b0625029906270238022702380227023802270238022a02350227029906270299062702990628023702260238022702380227023802270238022702380226023b02240299062702380226023802270238022602380227023c0223029906270299062702380226029b062402990627029906270299062802980627020000'))
        loop.run_until_complete(discoveryTest())
        loop.run_until_complete(emit_test('00000000ec000000000000000000dc00d2008604d200ec02d400e80ad4000d05d9000b05d2008404d4007103d4008132d400fc03d4001e06d400ec02d400ec02d4009705d2008404d400ec02d400ea02dc00ffff283e0100d200fd03d4008404d400ec02d400eb0ad2000d05d4000d05d4008404d4007103d6008132d200fd03d400610ad4002e07d400ec02d4009405d4008504d300ec02d400ea02d400ffff333e0100cc00fd03d4008404d300ed02d300eb0ad1000e05d3000e05d4008404d4007203d3008432d100fd03d300640ad1002f07d300ed02d3009505d3008504d400ee02d000eb02d3000000'))
        #loop.run_until_complete(learn_test())
    except BaseException as ex:
        _LOGGER.error("Test error %s",str(ex))
    except:
        _LOGGER.error("Test error %s",traceback.format_exc())
    finally:
        loop.close()


'''
Created on 25 apr 2019

@author: Matteo
'''
import time
from . import _LOGGER
from .const import (CD_ADD_AND_CONTINUE_WAITING,CD_RETURN_IMMEDIATELY,CD_CONTINUE_WAITING)
from .asyncio_udp import open_local_endpoint
import binascii
import struct
PORT = 10000
DISCOVERY_ALLONE = b'\x49\x52\x44'
DISCOVERY_S20 = b'\x53\x4f\x43'
MAC_START = b'\xac\xcf'
MAGIC = b'\x68\x64'
DISCOVERY_LEN = b'\x00\x06'
DISCOVERY_ID = b'\x71\x61'
SUBSCRIBE_LEN = b'\x00\x1e'
SUBSCRIBE_ID = b'\x63\x6c'
PADDING_1 = b'\x20\x20\x20\x20\x20\x20'
PADDING_2 = b'\x00\x00\x00\x00'
SUBSCRIPTION_TIMEOUT = 60

class OrviboUDP:
    _local = None
    def __repr__(self):
        return "[%s:%d] %s %s" % (*self.hp,OrviboUDP.print_mac(self.mac),time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(self.mytime)))
        
    def __init__(self,hp,mac,mytime = None,timeout=3,**kwargs):
        self.hp = hp
        if isinstance(mac, bytes):
            self.mac = mac
        else:
            mac = mac.replace(':','').replace('-','').replace(' ','')
            self.mac = binascii.unhexlify(mac)
        ba = bytearray(self.mac)
        ba.reverse()
        self.mac_reversed = bytes(ba)
        self.time_subscribe = 0
        self.mytime = mytime - 2208988800 if mytime is not None else time.time()
        self.learning_time = 0
        self.timeout = timeout
        
    def use_subscribe_data(self,s_data):
        pass
    
    @staticmethod
    async def protocol(data,addr,check_data_fun,timeout,retry=3,is_broadcast=False,**kwargs):
        out_data = None
        for _ in range(retry):
            try:
                if await OrviboUDP.init_local(**kwargs):
                    for _ in range(retry):
                        out_data = await OrviboUDP._local.protocol(data,addr,check_data_fun,timeout,1,is_broadcast)
                        if out_data:
                            break
                    break
            except BaseException as ex:
                OrviboUDP.destroy_local()
                _LOGGER.error("Protocol[%s:%d] error: %s",*addr,str(ex))
            except:
                OrviboUDP.destroy_local()
                _LOGGER.error("Protocol[%s:%d] error",*addr)
        #=======================================================================
        # if not out_data:
        #     OrviboUDP.destroy_local()
        #=======================================================================
        return out_data
    
    @staticmethod
    async def init_local(**kwargs):
        if not OrviboUDP._local:
            try:
                OrviboUDP._local = await open_local_endpoint(port=PORT,**kwargs)
            except BaseException as ex:
                _LOGGER.error("Open endpoint error %s",str(ex))
                OrviboUDP._local = None
            except:
                _LOGGER.error("Open endpoint error")
                OrviboUDP._local = None
        return OrviboUDP._local
    
    @staticmethod
    def destroy_local():
        if OrviboUDP._local:
            try:
                OrviboUDP._local.abort()
            except:
                pass
            OrviboUDP._local = None
        
    @staticmethod
    def print_mac(mac_bytes):
        return binascii.hexlify(mac_bytes).decode('utf-8')
    
    @staticmethod
    def check_discovery_packet(data,addr):
        return CD_CONTINUE_WAITING if len(data)<41 or data[4:6] != (DISCOVERY_ID) else CD_ADD_AND_CONTINUE_WAITING
        
    @staticmethod
    def mac_from_data(data):
        idx = data.find(MAC_START)
        if idx>=0 and idx+6<=len(data):
            return data[idx:idx+6]
        else:
            return None
        
    def is_my_mac(self,data):
        mac = OrviboUDP.mac_from_data(data)
        return False if not mac else mac==self.mac
        
    def check_subscription_packet(self,data,addr):
        return CD_RETURN_IMMEDIATELY if len(data)>=13 and data[4:6] == (SUBSCRIBE_ID) and self.is_my_mac(data) else CD_CONTINUE_WAITING
    
    async def subscribe_if_necessary(self,timeout=-1,retry=3):
        now = time.time()
        if now-self.time_subscribe>SUBSCRIPTION_TIMEOUT:
            timeout = self.timeout if timeout<=0 else timeout
            out_data = await OrviboUDP.protocol(MAGIC + SUBSCRIBE_LEN + SUBSCRIBE_ID + self.mac \
                                    + PADDING_1 + self.mac_reversed + PADDING_1,self.hp,self.check_subscription_packet,timeout,retry)
            if out_data:
                self.time_subscribe = time.time()
                self.use_subscribe_data(out_data[0])
                return True
            return False
        else:
            return True
        
    @staticmethod
    async def discovery(broadcast_address='255.255.255.255',timeout=5,retry=3):
        out_data = await OrviboUDP.protocol(MAGIC + DISCOVERY_LEN + DISCOVERY_ID,\
                              (broadcast_address,PORT),
                              OrviboUDP.check_discovery_packet, timeout, retry,is_broadcast=True,allow_broadcast=True)
        if out_data:
            hosts = dict()
            for d_a in out_data:
                data = d_a[0]
                keyv = '%s:%d' %  d_a[1]
                if keyv not in hosts:
                    if data.find(DISCOVERY_ALLONE)>=0:
                        tp = DISCOVERY_ALLONE
                    elif data.find(DISCOVERY_S20)>=0:
                        tp=DISCOVERY_S20
                    else:
                        _LOGGER.warning("Unknown device type %s %s", keyv,OrviboUDP.print_mac(data[31:37]))
                        continue
                    dev = dict(hp=d_a[1],type=tp,mac=data[7:13],mytime=struct.unpack('<I',data[37:41])[0],raw=data)
                    _LOGGER.info("Discovered device %s", dev)
                    hosts[keyv] = dev
            return hosts
        else:
            return dict()


""" taken from https://gist.github.com/vxgmichel/e47bff34b68adb3cf6bd4845c4bed448
Provide high-level UDP endpoints for asyncio.
Example:
async def main():
    # Create a local UDP enpoint
    local = await open_local_endpoint('localhost', 8888)
    # Create a remote UDP enpoint, pointing to the first one
    remote = await open_remote_endpoint(*local.address)
    # The remote endpoint sends a datagram
    remote.send(b'Hey Hey, My My')
    # The local endpoint receives the datagram, along with the address
    data, address = await local.receive()
    # This prints: Got 'Hey Hey, My My' from 127.0.0.1 port 8888
    print(f"Got {data!r} from {address[0]} port {address[1]}")
"""

import asyncio
import time
from . import _LOGGER
from .const import (CD_ADD_AND_CONTINUE_WAITING,CD_RETURN_IMMEDIATELY,CD_ABORT_AND_RETRY)


class DatagramEndpointProtocol(asyncio.DatagramProtocol):
    """Datagram protocol for the endpoint high-level interface."""

    def __init__(self, endpoint):
        self._endpoint = endpoint

    # Protocol methods

    def connection_made(self, transport):
        self._endpoint._transport = transport

    def connection_lost(self, exc):
        if exc is not None:  # pragma: no cover
            msg = 'Endpoint lost the connection: {!r}'
            _LOGGER.warning(msg.format(exc))
        self._endpoint.close()

    # Datagram protocol methods

    def datagram_received(self, data, addr):
        self._endpoint.feed_datagram(data, addr)

    def error_received(self, exc):
        msg = 'Endpoint received an error: {!r}'
        _LOGGER.warning(msg.format(exc))


# Enpoint classes

class Endpoint:
    """High-level interface for UDP enpoints.
    Can either be local or remote.
    It is initialized with an optional queue size for the incoming datagrams.
    """

    def __init__(self, queue_size=None):
        if queue_size is None:
            queue_size = 0
        self._queue = dict()
        self._closed = False
        self._transport = None
        self._broadcast = False
        self._queue_size = queue_size

    # Protocol callbacks

    def feed_datagram(self, data, addr):
        try:
            key = self._init_queue(addr)
            self._queue[key].put_nowait((data, addr))
        except asyncio.QueueFull:
            _LOGGER.warning('Endpoint[%s:%d] queue is full',*addr)

    def close(self):
        # Manage flag
        if self._closed:
            return
        self._closed = True
        # Wake up
        for a,q in self._queue.items():
            if q.empty():
                self.feed_datagram(None, a)
        # Close transport
        if self._transport:
            self._transport.close()

    # User methods
    
    async def protocol(self,data,addr,check_data_fun,timeout,retry=3,is_broadcast = False):
        lstdata = []
        if is_broadcast:
            self.broadcast = True
        for _ in range(retry):
            if data:
                self.send(data,addr,True)
            starttime = time.time()
            passed = 0
            while passed<timeout:
                try:
                    (rec_data,rec_addr) = await asyncio.wait_for(self.receive(addr), timeout-passed)
                    rv = check_data_fun(rec_data,rec_addr) 
                    if isinstance(rv, tuple):
                        rec_data = rv[1]
                        rv = rv[0]
                    if rv==CD_RETURN_IMMEDIATELY:
                        self.broadcast = False
                        return rec_data,rec_addr
                    elif rv==CD_ABORT_AND_RETRY:
                        break
                    elif rv==CD_ADD_AND_CONTINUE_WAITING:
                        lstdata.append((rec_data,rec_addr))
                except asyncio.TimeoutError:
                    _LOGGER.warning("Protocol[%s:%d] timeout",*addr)
                    break
                passed = time.time()-starttime
            if lstdata:
                self.broadcast = False
                return lstdata
            elif not data:
                break
        self.broadcast = False
        return None
    
    @property
    def broadcast(self):
        return self._broadcast
    @broadcast.setter
    def broadcast(self,v):
        self._broadcast = v
        
    
    def _init_queue(self,addr):
        if self._broadcast:
            key = '*'
        else:
            key = addr[0]
        if key not in self._queue:
            self._queue[key] =  asyncio.Queue(self._queue_size)
        return key

    def send(self, data, addr, expect_response):
        """Send a datagram to the given address."""
        if self._closed:
            raise IOError("Enpoint is closed")
        if expect_response:
            self._init_queue(addr)
        self._transport.sendto(data, addr)

    async def receive(self,expected_sender = ('*',0)):
        """Wait for an incoming datagram and return it with
        the corresponding address.
        This method is a coroutine.
        """
        key = self._init_queue(expected_sender)
        if self._queue[key].empty() and self._closed:
            raise IOError("Enpoint is closed")
        data, addr = await self._queue[key].get()
        if data is None:
            raise IOError("Enpoint is closed")
        return data, addr

    def abort(self):
        """Close the transport immediately."""
        if self._closed:
            raise IOError("Enpoint is closed")
        self._transport.abort()
        self.close()

    # Properties

    @property
    def address(self):
        """The endpoint address as a (host, port) tuple."""
        return self._transport._sock.getsockname()

    @property
    def closed(self):
        """Indicates whether the endpoint is closed or not."""
        return self._closed
# High-level coroutines

async def open_datagram_endpoint(
        host, port, *, endpoint_factory=Endpoint, remote=False, **kwargs):
    """Open and return a datagram endpoint.
    The default endpoint factory is the Endpoint class.
    The endpoint can be made local or remote using the remote argument.
    Extra keyword arguments are forwarded to `loop.create_datagram_endpoint`.
    """
    loop = asyncio.get_event_loop()
    endpoint = endpoint_factory()
    kwargs['remote_addr' if remote else 'local_addr'] = host, port
    kwargs['protocol_factory'] = lambda: DatagramEndpointProtocol(endpoint)
    await loop.create_datagram_endpoint(**kwargs,reuse_address=True, reuse_port=None)
    return endpoint


async def open_local_endpoint(
        host='0.0.0.0', port=0, *, queue_size=None, **kwargs):
    """Open and return a local datagram endpoint.
    An optional queue size arguement can be provided.
    Extra keyword arguments are forwarded to `loop.create_datagram_endpoint`.
    """
    return await open_datagram_endpoint(
        host, port, remote=False,
        endpoint_factory=lambda: Endpoint(queue_size),
        **kwargs)


async def open_remote_endpoint(
        host, port, *, queue_size=None, **kwargs):
    """Open and return a remote datagram endpoint.
    An optional queue size arguement can be provided.
    Extra keyword arguments are forwarded to `loop.create_datagram_endpoint`.
    """
    return await open_datagram_endpoint(
        host, port, remote=True,
        endpoint_factory=lambda: Endpoint(queue_size),
        **kwargs)



'''
Created on 25 apr 2019

@author: Matteo
'''
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="asyncio-orvibo",
    version="1.16",
    author="p3g4asus",
    author_email="fulminedipegasus@gmail.com",
    description="Asyncio module for Orvibo devices control",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/p3g4asus/asyncio-orvibo",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)